from __future__ import annotations

import math
from abc import ABC
from types import SimpleNamespace
from typing import Dict

from nmmo.datastore.serialized import SerializedState
from nmmo.lib.colors import Tier
from nmmo.lib.log import EventCode

# pylint: disable=no-member
ItemState = SerializedState.subclass("Item", [
  "id",
  "type_id",
  "owner_id",

  "level",
  "capacity",
  "quantity",
  "melee_attack",
  "range_attack",
  "mage_attack",
  "melee_defense",
  "range_defense",
  "mage_defense",
  "health_restore",
  "resource_restore",
  "equipped",

  # Market
  "listed_price",
])

# TODO: These limits should be defined in the config.
ItemState.Limits = lambda config: {
  "id": (0, math.inf),
  "type_id": (0, (config.ITEM_N + 1) if config.ITEM_SYSTEM_ENABLED else 0),
  "owner_id": (-math.inf, math.inf),
  "level": (0, 99),
  "capacity": (0, 99),
  "quantity": (0, math.inf), # NOTE: Ammunitions can be stacked infinitely
  "melee_attack": (0, 100),
  "range_attack": (0, 100),
  "mage_attack": (0, 100),
  "melee_defense": (0, 100),
  "range_defense": (0, 100),
  "mage_defense": (0, 100),
  "health_restore": (0, 100),
  "resource_restore": (0, 100),
  "equipped": (0, 1),
  "listed_price": (0, math.inf),
}

ItemState.Query = SimpleNamespace(
  table=lambda ds: ds.table("Item").where_neq(
    ItemState.State.attr_name_to_col["id"], 0),

  by_id=lambda ds, id: ds.table("Item").where_eq(
    ItemState.State.attr_name_to_col["id"], id),

  owned_by = lambda ds, id: ds.table("Item").where_eq(
    ItemState.State.attr_name_to_col["owner_id"], id),

  for_sale = lambda ds: ds.table("Item").where_neq(
    ItemState.State.attr_name_to_col["listed_price"], 0),
)

class Item(ItemState):
  ITEM_TYPE_ID = None
  _item_type_id_to_class: Dict[int, type] = {}

  @staticmethod
  def register(item_type):
    assert item_type.ITEM_TYPE_ID is not None
    if item_type.ITEM_TYPE_ID not in Item._item_type_id_to_class:
      Item._item_type_id_to_class[item_type.ITEM_TYPE_ID] = item_type

  @staticmethod
  def item_class(type_id: int):
    return Item._item_type_id_to_class[type_id]

  def __init__(self, realm, level,
              capacity=0,
              melee_attack=0, range_attack=0, mage_attack=0,
              melee_defense=0, range_defense=0, mage_defense=0,
              health_restore=0, resource_restore=0):

    super().__init__(realm.datastore, ItemState.Limits(realm.config))
    self.realm = realm
    self.config = realm.config

    Item.register(self.__class__)

    self.id.update(self.datastore_record.id)
    self.type_id.update(self.ITEM_TYPE_ID)
    self.level.update(level)
    self.capacity.update(capacity)
    # every item instance is created individually, i.e., quantity=1
    self.quantity.update(1)
    self.melee_attack.update(melee_attack)
    self.range_attack.update(range_attack)
    self.mage_attack.update(mage_attack)
    self.melee_defense.update(melee_defense)
    self.range_defense.update(range_defense)
    self.mage_defense.update(mage_defense)
    self.health_restore.update(health_restore)
    self.resource_restore.update(resource_restore)
    realm.items[self.id.val] = self

  def destroy(self):
    # NOTE: we may want to track the item lifecycle and
    #   and see how many high-level items are wasted
    if self.owner_id.val in self.realm.players:
      self.realm.players[self.owner_id.val].inventory.remove(self)
    self.realm.items.pop(self.id.val, None)
    self.datastore_record.delete()

  @property
  def packet(self):
    return {'item':             self.__class__.__name__,
            'level':            self.level.val,
            'capacity':         self.capacity.val,
            'quantity':         self.quantity.val,
            'melee_attack':     self.melee_attack.val,
            'range_attack':     self.range_attack.val,
            'mage_attack':      self.mage_attack.val,
            'melee_defense':    self.melee_defense.val,
            'range_defense':    self.range_defense.val,
            'mage_defense':     self.mage_defense.val,
            'health_restore':   self.health_restore.val,
            'resource_restore': self.resource_restore.val,
            }

  def _level(self, entity):
    # this is for armors, ration, and potion
    # weapons and tools must override this with specific skills
    return entity.level

  def level_gt(self, entity):
    return self.level.val > self._level(entity)

  def use(self, entity) -> bool:
    raise NotImplementedError

class Stack:
  @property
  def signature(self):
    return (self.type_id.val, self.level.val)

class Equipment(Item):
  @property
  def packet(self):
    packet = {'color': self.color.packet()}
    return {**packet, **super().packet}

  @property
  def color(self):
    if self.level == 0:
      return Tier.BLACK
    if self.level < 10:
      return Tier.WOOD
    if self.level < 20:
      return Tier.BRONZE
    if self.level < 40:
      return Tier.SILVER
    if self.level < 60:
      return Tier.GOLD
    if self.level < 80:
      return Tier.PLATINUM
    return Tier.DIAMOND

  def unequip(self, equip_slot):
    assert self.equipped.val == 1
    self.equipped.update(0)
    equip_slot.unequip()

  def equip(self, entity, equip_slot):
    assert self.equipped.val == 0
    if self._level(entity) < self.level.val:
      return

    self.equipped.update(1)
    equip_slot.equip(self)

    if self.config.LOG_MILESTONES and entity.is_player and self.config.LOG_VERBOSE:
      for (label, level) in [
        (f"{self.__class__.__name__}_Level", self.level.val),
        ("Item_Level", entity.equipment.item_level),
        ("Melee_Attack", entity.equipment.melee_attack),
        ("Range_Attack", entity.equipment.range_attack),
        ("Mage_Attack", entity.equipment.mage_attack),
        ("Melee_Defense", entity.equipment.melee_defense),
        ("Range_Defense", entity.equipment.range_defense),
        ("Mage_Defense", entity.equipment.mage_defense)]:

        self.realm.log_milestone(label, level, f'EQUIPMENT: {label} {level}')

  def _slot(self, entity):
    raise NotImplementedError

  def use(self, entity):
    assert self in entity.inventory, "Item is not in entity's inventory"
    assert self.listed_price == 0, "Listed item cannot be used"
    assert self._level(entity) >= self.level.val, "Entity's level is not sufficient to use the item"

    if self.equipped.val:
      self.unequip(self._slot(entity))
    else:
      # always empty the slot first
      self._slot(entity).unequip()
      self.equip(entity, self._slot(entity))
      self.realm.event_log.record(EventCode.EQUIP_ITEM, entity, item=self)

class Armor(Equipment, ABC):
  def __init__(self, realm, level, **kwargs):
    defense = realm.config.EQUIPMENT_ARMOR_BASE_DEFENSE + \
              level*realm.config.EQUIPMENT_ARMOR_LEVEL_DEFENSE
    super().__init__(realm, level,
                     melee_defense=defense,
                     range_defense=defense,
                     mage_defense=defense,
                     **kwargs)
class Hat(Armor):
  ITEM_TYPE_ID = 2
  def _slot(self, entity):
    return entity.inventory.equipment.hat
class Top(Armor):
  ITEM_TYPE_ID = 3
  def _slot(self, entity):
    return entity.inventory.equipment.top
class Bottom(Armor):
  ITEM_TYPE_ID = 4
  def _slot(self, entity):
    return entity.inventory.equipment.bottom


class Weapon(Equipment):
  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.attack = (
      realm.config.EQUIPMENT_WEAPON_BASE_DAMAGE +
      level*realm.config.EQUIPMENT_WEAPON_LEVEL_DAMAGE)

  def _slot(self, entity):
    return entity.inventory.equipment.held

class Spear(Weapon):
  ITEM_TYPE_ID = 5

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.melee_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.melee.level.val
class Bow(Weapon):
  ITEM_TYPE_ID = 6

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.range_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.range.level.val
class Wand(Weapon):
  ITEM_TYPE_ID = 7

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.mage_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.mage.level.val


class Tool(Equipment):
  def __init__(self, realm, level, **kwargs):
    defense = realm.config.EQUIPMENT_TOOL_BASE_DEFENSE + \
        level*realm.config.EQUIPMENT_TOOL_LEVEL_DEFENSE
    super().__init__(realm, level,
                      melee_defense=defense,
                      range_defense=defense,
                      mage_defense=defense,
                      **kwargs)

  def _slot(self, entity):
    return entity.inventory.equipment.held
class Rod(Tool):
  ITEM_TYPE_ID = 8
  def _level(self, entity):
    return entity.skills.fishing.level.val
class Gloves(Tool):
  ITEM_TYPE_ID = 9
  def _level(self, entity):
    return entity.skills.herbalism.level.val
class Pickaxe(Tool):
  ITEM_TYPE_ID = 10
  def _level(self, entity):
    return entity.skills.prospecting.level.val
class Axe(Tool):
  ITEM_TYPE_ID = 11
  def _level(self, entity):
    return entity.skills.carving.level.val
class Chisel(Tool):
  ITEM_TYPE_ID = 12
  def _level(self, entity):
    return entity.skills.alchemy.level.val


class Ammunition(Equipment, Stack):
  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.attack = (
      realm.config.EQUIPMENT_AMMUNITION_BASE_DAMAGE +
      level*realm.config.EQUIPMENT_AMMUNITION_LEVEL_DAMAGE)

  def _slot(self, entity):
    return entity.inventory.equipment.ammunition

  def fire(self, entity) -> int:
    assert self.equipped.val > 0, 'Ammunition not equipped'
    assert self.quantity.val > 0, 'Used ammunition with 0 quantity'

    self.quantity.decrement()

    if self.quantity.val == 0:
      entity.inventory.remove(self)
      # delete this empty item instance from the datastore
      self.destroy()

    return self.damage

class Whetstone(Ammunition):
  ITEM_TYPE_ID = 13

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.melee_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.melee.level.val

  @property
  def damage(self):
    return self.melee_attack.val

class Arrow(Ammunition):
  ITEM_TYPE_ID = 14

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.range_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.range.level.val

  @property
  def damage(self):
    return self.range_attack.val

class Runes(Ammunition):
  ITEM_TYPE_ID = 15

  def __init__(self, realm, level, **kwargs):
    super().__init__(realm, level, **kwargs)
    self.mage_attack.update(self.attack)

  def _level(self, entity):
    return entity.skills.mage.level.val

  @property
  def damage(self):
    return self.mage_attack.val


# NOTE: Each consumable item (ration, potion) cannot be stacked,
#   so each item takes 1 inventory space
class Consumable(Item):
  def use(self, entity) -> bool:
    assert self in entity.inventory, "Item is not in entity's inventory"
    assert self.listed_price == 0, "Listed item cannot be used"
    assert self._level(entity) >= self.level.val, "Entity's level is not sufficient to use the item"

    self.realm.log_milestone(
      f'Consumed_{self.__class__.__name__}', self.level.val,
      f"PROF: Consumed {self.level.val} {self.__class__.__name__} "
      f"by Entity level {entity.attack_level}",
      tags={"player_id": entity.ent_id})

    self.realm.event_log.record(EventCode.CONSUME_ITEM, entity, item=self)

    self._apply_effects(entity)
    entity.inventory.remove(self)
    self.destroy()
    return True

class Ration(Consumable):
  ITEM_TYPE_ID = 16

  def __init__(self, realm, level, **kwargs):
    restore = 0
    if realm.config.PROFESSION_SYSTEM_ENABLED:
      restore = realm.config.PROFESSION_CONSUMABLE_RESTORE(level)
    super().__init__(realm, level, resource_restore=restore, **kwargs)

  def _apply_effects(self, entity):
    entity.resources.food.increment(self.resource_restore.val)
    entity.resources.water.increment(self.resource_restore.val)

class Potion(Consumable):
  ITEM_TYPE_ID = 17

  def __init__(self, realm, level, **kwargs):
    restore = 0
    if realm.config.PROFESSION_SYSTEM_ENABLED:
      restore = realm.config.PROFESSION_CONSUMABLE_RESTORE(level)
    super().__init__(realm, level, health_restore=restore, **kwargs)

  def _apply_effects(self, entity):
    entity.resources.health.increment(self.health_restore.val)
    entity.poultice_consumed += 1
    entity.poultice_level_consumed = max(
      entity.poultice_level_consumed, self.level.val)



# import time
import cProfile
import io
import pstats

import nmmo
from nmmo.core.config import (NPC, AllGameSystems, Combat, Communication,
                              Equipment, Exchange, Item, Medium, Profession,
                              Progression, Resource, Small, Terrain)
from nmmo.task.task_api import nmmo_default_task, make_same_task
from nmmo.task.base_predicates import CountEvent, FullyArmed
from nmmo.systems.skill import Melee
from tests.testhelpers import profile_env_step
from scripted import baselines


# Test utils
def create_and_reset(conf):
  env = nmmo.Env(conf())
  env.reset(map_id=1)

def create_config(base, *systems):
  systems   = (base, *systems)
  name      = '_'.join(cls.__name__ for cls in systems)

  conf                    = type(name, systems, {})()

  conf.TERRAIN_TRAIN_MAPS = 1
  conf.TERRAIN_EVAL_MAPS  = 1
  conf.IMMORTAL = True

  return conf

def benchmark_config(benchmark, base, nent, *systems):
  conf = create_config(base, *systems)
  conf.PLAYER_N = nent
  conf.PLAYERS = [baselines.Random]

  env = nmmo.Env(conf)
  env.reset()

  benchmark(env.step, actions={})

# Small map tests -- fast with greater coverage for individual game systems
def test_small_env_creation(benchmark):
  benchmark(lambda: nmmo.Env(Small()))

def test_small_env_reset(benchmark):
  config = Small()
  config.PLAYERS = [baselines.Random]
  env = nmmo.Env(config)
  benchmark(lambda: env.reset(map_id=1))

def test_fps_base_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1)

def test_fps_minimal_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1, Terrain, Resource, Combat, Progression)

def test_fps_npc_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1, Terrain, Resource, Combat, Progression, NPC)

def test_fps_test_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1, Terrain, Resource, Combat, Progression, Item, Exchange)

def test_fps_no_npc_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1, Terrain, Resource,
    Combat, Progression, Item, Equipment, Profession, Exchange, Communication)

def test_fps_all_small_1_pop(benchmark):
  benchmark_config(benchmark, Small, 1, AllGameSystems)

def test_fps_base_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1)

def test_fps_minimal_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1, Terrain, Resource, Combat)

def test_fps_npc_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1, Terrain, Resource, Combat, NPC)

def test_fps_test_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1, Terrain, Resource, Combat, Progression, Item, Exchange)

def test_fps_no_npc_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1, Terrain, Resource,
    Combat, Progression, Item, Equipment, Profession, Exchange, Communication)

def test_fps_all_med_1_pop(benchmark):
  benchmark_config(benchmark, Medium, 1, AllGameSystems)

def test_fps_base_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100)

def test_fps_minimal_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100, Terrain, Resource, Combat)

def test_fps_npc_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100, Terrain, Resource, Combat, NPC)

def test_fps_test_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100, Terrain, Resource, Combat, Progression, Item, Exchange)

def test_fps_no_npc_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100, Terrain, Resource, Combat,
    Progression, Item, Equipment, Profession, Exchange, Communication)

def test_fps_all_med_100_pop(benchmark):
  benchmark_config(benchmark, Medium, 100, AllGameSystems)

def set_seed_test():
  random_seed = 5000
  conf = create_config(Medium, Terrain, Resource, Combat, NPC)
  conf.PLAYER_N = 10
  conf.PLAYERS = [baselines.Random]

  env = nmmo.Env(conf)

  env.reset(seed=random_seed)
  for _ in range(1024):
    env.step({})

def set_seed_test_complex():
  tasks = nmmo_default_task(range(128))
  tasks += make_same_task(CountEvent, range(128),
                          pred_kwargs={'event': 'EAT_FOOD', 'N': 10})
  tasks += make_same_task(FullyArmed, range(128),
                          pred_kwargs={'combat_style': Melee, 'level': 3, 'num_agent': 1})
  profile_env_step(tasks=tasks)

if __name__ == '__main__':
  with open('profile.run','a', encoding="utf-8") as f:
    pr = cProfile.Profile()
    pr.enable()
    set_seed_test_complex()
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr,stream=s).sort_stats('tottime')
    ps.print_stats()
    f.write(s.getvalue())

'''
def benchmark_env(benchmark, env, nent):
  env.config.PLAYER_N = nent
  env.config.PLAYERS = [nmmo.agent.Random]
  env.reset()

  benchmark(env.step, actions={})
# Reuse large maps since we aren't benchmarking the reset function
def test_large_env_creation(benchmark):
  benchmark(lambda: nmmo.Env(Large()))

def test_large_env_reset(benchmark):
  env = nmmo.Env(Large())
  benchmark(lambda: env.reset(idx=1))

LargeMapsRCP = nmmo.Env(create_config(Large, Resource, Terrain, Combat, Progression))
LargeMapsAll = nmmo.Env(create_config(Large, AllGameSystems))

def test_fps_large_rcp_1_pop(benchmark):
  benchmark_env(benchmark, LargeMapsRCP, 1)

def test_fps_large_rcp_100_pop(benchmark):
  benchmark_env(benchmark, LargeMapsRCP, 100)

def test_fps_large_rcp_1000_pop(benchmark):
  benchmark_env(benchmark, LargeMapsRCP, 1000)

def test_fps_large_all_1_pop(benchmark):
  benchmark_env(benchmark, LargeMapsAll, 1)

def test_fps_large_all_100_pop(benchmark):
  benchmark_env(benchmark, LargeMapsAll, 100)

def test_fps_large_all_1000_pop(benchmark):
  benchmark_env(benchmark, LargeMapsAll, 1000)
'''



from __future__ import annotations
from ast import Tuple

import math
from types import SimpleNamespace
from typing import Dict, List
from nmmo.datastore.datastore import Datastore, DatastoreRecord

"""
This code defines classes for serializing and deserializing data
in a structured way.

The SerializedAttribute class represents a single attribute of a
record and provides methods for updating and querying its value,
as well as enforcing minimum and maximum bounds on the value.

The SerializedState class serves as a base class for creating
serialized representations of specific types of data, using a
list of attribute names to define the structure of the data.
The subclass method is a factory method for creating subclasses
of SerializedState that are tailored to specific types of data.
"""

class SerializedAttribute():
  def __init__(self,
      name: str,
      datastore_record: DatastoreRecord,
      column: int, min_val=-math.inf, max_val=math.inf) -> None:
    self._name = name
    self.datastore_record = datastore_record
    self._column = column
    self._min = min_val
    self._max = max_val
    self._val = 0

  @property
  def val(self):
    return self._val

  def update(self, value):
    if value > self._max:
      value = self._max
    elif value < self._min:
      value = self._min
    self.datastore_record.update(self._column, value)
    self._val = value

  @property
  def min(self):
    return self._min

  @property
  def max(self):
    return self._max

  def increment(self, val=1, max_v=math.inf):
    self.update(min(max_v, self.val + val))
    return self

  def decrement(self, val=1, min_v=-math.inf):
    self.update(max(min_v, self.val - val))
    return self

  @property
  def empty(self):
    return self.val == 0

  def __eq__(self, other):
    return self.val == other

  def __ne__(self, other):
    return self.val != other

  def __lt__(self, other):
    return self.val < other

  def __le__(self, other):
    return self.val <= other

  def __gt__(self, other):
    return self.val > other

  def __ge__(self, other):
    return self.val >= other

class SerializedState():
  @staticmethod
  def subclass(name: str, attributes: List[str]):
    class Subclass(SerializedState):
      _name = name
      State = SimpleNamespace(
        attr_name_to_col = {a: i for i, a in enumerate(attributes)},
        num_attributes = len(attributes),
        table = lambda ds: ds.table(name)
      )

      def __init__(self, datastore: Datastore,
                   limits: Dict[str, Tuple[float, float]] = None):

        limits = limits or {}
        self.datastore_record = datastore.create_record(name)

        for attr, col in self.State.attr_name_to_col.items():
          try:
            setattr(self, attr,
              SerializedAttribute(attr, self.datastore_record, col,
                *limits.get(attr, (-math.inf, math.inf))))
          except Exception as exc:
            raise RuntimeError('Failed to set attribute' + attr) from exc

      @classmethod
      def parse_array(cls, data) -> SimpleNamespace:
        # Takes in a data array and returns a SimpleNamespace object with
        # attribute names as keys and corresponding values from the input
        # data array.
        assert len(data) == cls.State.num_attributes, \
          f"Expected {cls.State.num_attributes} attributes, got {len(data)}"
        return SimpleNamespace(**{
          attr: data[col] for attr, col in cls.State.attr_name_to_col.items()
        })

    return Subclass



#pylint: disable=invalid-name, unused-argument, no-value-for-parameter
from __future__ import annotations
from typing import Iterable
import numpy as np
from numpy import count_nonzero as count

from nmmo.task.group import Group
from nmmo.task.game_state import GameState
from nmmo.systems import skill as nmmo_skill
from nmmo.systems.skill import Skill
from nmmo.systems.item import Item
from nmmo.lib.material import Material
from nmmo.lib import utils

def norm(progress):
  return max(min(progress, 1.0), 0.0)

def Success(gs: GameState, subject: Group):
  ''' Returns True. For debugging.
  '''
  return True

def TickGE(gs: GameState, subject: Group, num_tick: int):
  """True if the current tick is greater than or equal to the specified num_tick.
  Is progress counter.
  """
  return norm(gs.current_tick / num_tick)

def CanSeeTile(gs: GameState, subject: Group, tile_type: type[Material]):
  """ True if any agent in subject can see a tile of tile_type
  """
  return any(tile_type.index in t for t in subject.obs.tile.material_id)

def StayAlive(gs: GameState, subject: Group):
  """True if all subjects are alive.
  """
  return count(subject.health > 0) == len(subject)

def AllDead(gs: GameState, subject: Group):
  """True if all subjects are dead.
  """
  return norm(1.0 - count(subject.health) / len(subject))

def OccupyTile(gs: GameState, subject: Group, row: int, col: int):
  """True if any subject agent is on the desginated tile.
  """
  return np.any((subject.row == row) & (subject.col == col))

def AllMembersWithinRange(gs: GameState, subject: Group, dist: int):
  """True if the max l-inf distance of teammates is
         less than or equal to dist
  """
  current_dist = max(subject.row.max()-subject.row.min(),
      subject.col.max()-subject.col.min())
  if current_dist <= 0:
    return 1.0
  return norm(dist / current_dist)

def CanSeeAgent(gs: GameState, subject: Group, target: int):
  """True if obj_agent is present in the subjects' entities obs.
  """
  return any(target in e.ids for e in subject.obs.entities)

def CanSeeGroup(gs: GameState, subject: Group, target: Iterable[int]):
  """ Returns True if subject can see any of target
  """
  if target is None:
    return False
  return any(CanSeeAgent(gs, subject, agent) for agent in target)

def DistanceTraveled(gs: GameState, subject: Group, dist: int):
  """True if the summed l-inf distance between each agent's current pos and spawn pos
        is greater than or equal to the specified _dist.
  """
  if not any(subject.health > 0):
    return False
  r = subject.row
  c = subject.col
  dists = utils.linf(list(zip(r,c)),[gs.spawn_pos[id_] for id_ in subject.entity.id])
  return norm(dists.sum() / dist)

def AttainSkill(gs: GameState, subject: Group,
                skill: type[Skill], level: int, num_agent: int):
  """True if the number of agents having skill level GE level
        is greather than or equal to num_agent
  """
  if level <= 1:
    return 1.0
  skill_level = getattr(subject,skill.__name__.lower() + '_level') - 1  # base level is 1
  return norm(sum(skill_level) / (num_agent * (level-1)))

def GainExperience(gs: GameState, subject: Group,
                   skill: type[Skill], experience: int, num_agent: int):
  """True if the experience gained for the skill is greater than or equal to experience."""
  skill_exp = getattr(subject,skill.__name__.lower() + '_exp')
  return norm(sum(skill_exp) / (experience*num_agent))

def CountEvent(gs: GameState, subject: Group, event: str, N: int):
  """True if the number of events occured in subject corresponding
      to event >= N
  """
  return norm(len(getattr(subject.event, event)) / N)

def ScoreHit(gs: GameState, subject: Group, combat_style: type[Skill], N: int):
  """True if the number of hits scored in style
  combat_style >= count
  """
  hits = subject.event.SCORE_HIT.combat_style == combat_style.SKILL_ID
  return norm(count(hits) / N)

def DefeatEntity(gs: GameState, subject: Group, agent_type: str, level: int, num_agent: int):
  """True if the number of agents (agent_type, >= level) defeated
        is greater than or equal to num_agent
  """
  # NOTE: there is no way to tell if an agent is a teammate or an enemy
  #   so agents can get rewarded for killing their own teammates
  defeated_type = subject.event.PLAYER_KILL.target_ent > 0 if agent_type == 'player' \
                    else subject.event.PLAYER_KILL.target_ent < 0
  defeated = defeated_type & (subject.event.PLAYER_KILL.level >= level)
  if num_agent > 0:
    return norm(count(defeated) / num_agent)
  return 1.0

def HoardGold(gs: GameState, subject: Group, amount: int):
  """True iff the summed gold of all teammate is greater than or equal to amount.
  """
  return norm(subject.gold.sum() / amount)

def EarnGold(gs: GameState, subject: Group, amount: int):
  """ True if the total amount of gold earned is greater than or equal to amount.
  """
  return norm(subject.event.EARN_GOLD.gold.sum() / amount)

def SpendGold(gs: GameState, subject: Group, amount: int):
  """ True if the total amount of gold spent is greater than or equal to amount.
  """
  return norm(subject.event.BUY_ITEM.gold.sum() / amount)

def MakeProfit(gs: GameState, subject: Group, amount: int):
  """ True if the total amount of gold earned-spent is greater than or equal to amount.
  """
  profits = subject.event.EARN_GOLD.gold.sum()
  costs = subject.event.BUY_ITEM.gold.sum()
  return  norm((profits-costs) / amount)

def InventorySpaceGE(gs: GameState, subject: Group, space: int):
  """True if the inventory space of every subjects is greater than or equal to
       the space. Otherwise false.
  """
  max_space = gs.config.ITEM_INVENTORY_CAPACITY
  return all(max_space - inv.len >= space for inv in subject.obs.inventory)

def OwnItem(gs: GameState, subject: Group, item: type[Item], level: int, quantity: int):
  """True if the number of items owned (_item_type, >= level)
     is greater than or equal to quantity.
  """
  owned = (subject.item.type_id == item.ITEM_TYPE_ID) & \
          (subject.item.level >= level)
  return norm(sum(subject.item.quantity[owned]) / quantity)

def EquipItem(gs: GameState, subject: Group, item: type[Item], level: int, num_agent: int):
  """True if the number of agents that equip the item (_item_type, >=_level)
     is greater than or equal to _num_agent.
  """
  equipped = (subject.item.type_id == item.ITEM_TYPE_ID) & \
             (subject.item.level >= level) & \
             (subject.item.equipped > 0)
  if num_agent > 0:
    return norm(count(equipped) / num_agent)
  return 1.0

def FullyArmed(gs: GameState, subject: Group,
               combat_style: type[Skill], level: int, num_agent: int):
  """True if the number of fully equipped agents is greater than or equal to _num_agent
       Otherwise false.
       To determine fully equipped, we look at hat, top, bottom, weapon, ammo, respectively,
       and see whether these are equipped and has level greater than or equal to _level.
  """
  WEAPON_IDS = {
    nmmo_skill.Melee: {'weapon':5, 'ammo':13}, # Spear, Whetstone
    nmmo_skill.Range: {'weapon':6, 'ammo':14}, # Bow, Arrow
    nmmo_skill.Mage: {'weapon':7, 'ammo':15} # Wand, Runes
  }
  item_ids = { 'hat':2, 'top':3, 'bottom':4 }
  item_ids.update(WEAPON_IDS[combat_style])

  lvl_flt = (subject.item.level >= level) & \
            (subject.item.equipped > 0)
  type_flt = np.isin(subject.item.type_id,list(item_ids.values()))
  _, equipment_numbers = np.unique(subject.item.owner_id[lvl_flt & type_flt],
                                   return_counts=True)
  if num_agent > 0:
    return norm((equipment_numbers >= len(item_ids.items())).sum() / num_agent)
  return 1.0

def ConsumeItem(gs: GameState, subject: Group, item: type[Item], level: int, quantity: int):
  """True if total quantity consumed of item type above level is >= quantity
  """
  type_flt = subject.event.CONSUME_ITEM.type == item.ITEM_TYPE_ID
  lvl_flt = subject.event.CONSUME_ITEM.level >= level
  return norm(subject.event.CONSUME_ITEM.number[type_flt & lvl_flt].sum() / quantity)

def HarvestItem(gs: GameState, subject: Group, item: type[Item], level: int, quantity: int):
  """True if total quantity harvested of item type above level is >= quantity
  """
  type_flt = subject.event.HARVEST_ITEM.type == item.ITEM_TYPE_ID
  lvl_flt = subject.event.HARVEST_ITEM.level >= level
  return norm(subject.event.HARVEST_ITEM.number[type_flt & lvl_flt].sum() / quantity)

def ListItem(gs: GameState, subject: Group, item: type[Item], level: int, quantity: int):
  """True if total quantity listed of item type above level is >= quantity
  """
  type_flt = subject.event.LIST_ITEM.type == item.ITEM_TYPE_ID
  lvl_flt = subject.event.LIST_ITEM.level >= level
  return norm(subject.event.LIST_ITEM.number[type_flt & lvl_flt].sum() / quantity)

def BuyItem(gs: GameState, subject: Group, item: type[Item], level: int, quantity: int):
  """True if total quantity purchased of item type above level is >= quantity
  """
  type_flt = subject.event.BUY_ITEM.type == item.ITEM_TYPE_ID
  lvl_flt = subject.event.BUY_ITEM.level >= level
  return norm(subject.event.BUY_ITEM.number[type_flt & lvl_flt].sum() / quantity)



from itertools import chain

from setuptools import find_packages, setup

REPO_URL = "https://github.com/neuralmmo/environment"

extra = {
    'docs': [
        'sphinx==5.0.0',
        'sphinx-rtd-theme==0.5.1',
        'sphinxcontrib-youtube==1.0.1',
        'myst-parser==1.0.0',
        'sphinx-rtd-theme==0.5.1',
        'sphinx-design==0.4.1',
        'furo==2023.3.27',
    ],
}

extra['all'] = list(set(chain.from_iterable(extra.values())))

with open('nmmo/version.py', encoding="utf-8") as vf:
  ver = vf.read().split()[-1].strip("'")

setup(
  name="nmmo",
  description="Neural MMO is a platform for multiagent intelligence research " + \
              "inspired by Massively Multiplayer Online (MMO) role-playing games. " + \
              "Documentation hosted at neuralmmo.github.io.",
  long_description_content_type="text/markdown",
  version=ver,
  packages=find_packages(),
  include_package_data=True,
  install_requires=[
    'numpy==1.23.3',
    'scipy==1.10.0',
    'pytest==7.3.0',
    'pytest-benchmark==3.4.1',
    # Deprecated renderer breaks py311
    #'autobahn==19.3.3',
    #'Twisted==19.2.0',
    'vec-noise==1.1.4',
    'imageio==2.23.0',
    'ordered-set==4.1.0',
    'pettingzoo>=1.19.0',
    'gym==0.23.0',
    'pylint==2.16.0',
    'psutil==5.9.3',
    'py==1.11.0',
    'tqdm<5',
    'dill==0.3.6',
  ],
  extras_require=extra,
  python_requires=">=3.7",
  license="MIT",
  author="Joseph Suarez",
  author_email="jsuarez@mit.edu",
  url=REPO_URL,
  keywords=["Neural MMO", "MMO"],
  classifiers=[
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Environment :: Console",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
  ],
)



from typing import List

import numpy as np

from nmmo.datastore.datastore import Datastore, DataTable


class NumpyTable(DataTable):
  def __init__(self, num_columns: int, initial_size: int, dtype=np.int16):
    super().__init__(num_columns)
    self._dtype  = dtype
    self._initial_size = initial_size
    self._max_rows = 0
    self._data = np.zeros((0, self._num_columns), dtype=self._dtype)
    self._expand(self._initial_size)

  def reset(self):
    super().reset() # resetting _id_allocator
    self._max_rows = 0
    self._data = np.zeros((0, self._num_columns), dtype=self._dtype)
    self._expand(self._initial_size)

  def update(self, row_id: int, col: int, value):
    self._data[row_id, col] = value

  def get(self, ids: List[int]):
    return self._data[ids]

  def where_eq(self, col: int, value):
    return self._data[self._data[:,col] == value]

  def where_neq(self, col: int, value):
    return self._data[self._data[:,col] != value]

  def where_in(self, col: int, values: List):
    return self._data[np.isin(self._data[:,col], values)]

  def window(self, row_idx: int, col_idx: int, row: int, col: int, radius: int):
    return self._data[(
      (np.abs(self._data[:,row_idx] - row) <= radius) &
      (np.abs(self._data[:,col_idx] - col) <= radius)
    ).ravel()]

  def add_row(self) -> int:
    if self._id_allocator.full():
      self._expand(self._max_rows * 2)
    row_id = self._id_allocator.allocate()
    return row_id

  def remove_row(self, row_id: int) -> int:
    self._id_allocator.remove(row_id)
    self._data[row_id] = 0

  def _expand(self, max_rows: int):
    assert max_rows > self._max_rows
    data = np.zeros((max_rows, self._num_columns), dtype=self._dtype)
    data[:self._max_rows] = self._data
    self._max_rows = max_rows
    self._id_allocator.expand(max_rows)
    self._data = data

  def is_empty(self) -> bool:
    all_data_zero = np.sum(self._data)==0
    # 0th row is reserved as padding, so # of free ids is _max_rows-1
    all_id_free = len(self._id_allocator.free) == self._max_rows-1
    return all_data_zero and all_id_free

class NumpyDatastore(Datastore):
  def _create_table(self, num_columns: int) -> DataTable:
    return NumpyTable(num_columns, 100)



import os
import json
import logging
import lzma
import pickle
from typing import Dict

from .render_utils import np_encoder, patch_packet

class ReplayHelper:
  def __init__(self):
    self._realm = None

  def set_realm(self, realm) -> None:
    self._realm = realm

  def reset(self):
    pass

  def update(self):
    pass

  def save(self, filename_prefix, compress):
    pass

class DummyReplayHelper(ReplayHelper):
  pass

class FileReplayHelper(ReplayHelper):
  def __init__(self, realm=None):
    super().__init__()
    self._realm = realm
    self.packets = None
    self.map = None
    self._i = 0

  def reset(self):
    self.packets = []
    self.map = None
    self._i = 0
    self.update() # to capture the initial packet

  def __len__(self):
    return len(self.packets)

  def __iter__(self):
    self._i = 0
    return self

  def __next__(self):
    if self._i >= len(self.packets):
      raise StopIteration
    packet = self.packets[self._i]
    packet['environment'] = self.map
    self._i += 1
    return packet

  def _packet(self):
    assert self._realm is not None, 'Realm not set'

    # TODO: remove patch_packet
    packet = patch_packet(self._realm.packet(), self._realm)

    if "environment" in packet:
      self.map = packet["environment"]
      del packet["environment"]
    if "config" in packet:
      del packet["config"]

    return packet

  def _metadata(self) -> Dict:
    return {
      'event_log': self._realm.event_log.get_data(),
      'event_attr_col': self._realm.event_log.attr_to_col
    }

  def update(self):
    self.packets.append(self._packet())

  def save(self, filename_prefix, compress=False):
    replay_file = f'{filename_prefix}.replay.json'
    metadata_file = f'{filename_prefix}.metadata.pkl'

    data = json.dumps({
      'map': self.map,
      'packets': self.packets
    }, default=np_encoder).encode('utf8')

    if compress:
      replay_file = f'{filename_prefix}.replay.lzma'
      data = lzma.compress(data, format=lzma.FORMAT_ALONE)

    logging.info('Saving replay to %s ...', replay_file)

    with open(replay_file, 'wb') as out:
      out.write(data)

    with open(metadata_file, 'wb') as out:
      pickle.dump(self._metadata(), out)

  @classmethod
  def load(cls, replay_file):
    extension = os.path.splitext(replay_file)[1]
    with open(replay_file, 'rb') as fp:
      data = fp.read()

    if extension != '.json':
      data = lzma.decompress(data, format=lzma.FORMAT_ALONE)
    data = json.loads(data.decode('utf-8'))

    replay_helper = FileReplayHelper()
    replay_helper.map = data['map']
    replay_helper.packets = data['packets']

    return replay_helper



#pylint: disable=protected-access, invalid-name

import heapq
from typing import Tuple

import numpy as np

from nmmo.lib.utils import in_bounds


def validTarget(ent, targ, rng):
  if targ is None or not targ.alive or lInfty(ent.pos, targ.pos) > rng:
    return False
  return True


def validResource(ent, tile, rng):
  return tile is not None and tile.state.tex in (
    'foilage', 'water') and lInfty(ent.pos, tile.pos) <= rng


def directionTowards(ent, targ):
  sr, sc = ent.pos
  tr, tc = targ.pos

  if abs(sc - tc) > abs(sr - tr):
    direction = (0, np.sign(tc - sc))
  else:
    direction = (np.sign(tr - sr), 0)

  return direction


def closestTarget(ent, tiles, rng=1):
  sr, sc = ent.pos
  for d in range(rng+1):
    for r in range(-d, d+1):
      for e in tiles[sr+r, sc-d].entities.values():
        if e is not ent and validTarget(ent, e, rng):
          return e

      for e in tiles[sr + r, sc + d].entities.values():
        if e is not ent and validTarget(ent, e, rng):
          return e

      for e in tiles[sr - d, sc + r].entities.values():
        if e is not ent and validTarget(ent, e, rng):
          return e

      for e in tiles[sr + d, sc + r].entities.values():
        if e is not ent and validTarget(ent, e, rng):
          return e
  return None


def lInf(ent, targ):
  sr, sc = ent.pos
  gr, gc = targ.pos
  return abs(gr - sr) + abs(gc - sc)


def adjacentPos(pos):
  r, c = pos
  return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1)]


def cropTilesAround(position: Tuple[int, int], horizon: int, tiles):
  line, column = position

  return tiles[max(line - horizon, 0): min(line + horizon + 1, len(tiles)),
                max(column - horizon, 0): min(column + horizon + 1, len(tiles[0]))]

# A* Search


def l1(start, goal):
  sr, sc = start
  gr, gc = goal
  return abs(gr - sr) + abs(gc - sc)


def l2(start, goal):
  sr, sc = start
  gr, gc = goal
  return 0.5*((gr - sr)**2 + (gc - sc)**2)**0.5

# TODO: unify lInfty and lInf


def lInfty(start, goal):
  sr, sc = start
  gr, gc = goal
  return max(abs(gr - sr), abs(gc - sc))


CUTOFF = 100


def aStar(realm_map, start, goal):
  cutoff = CUTOFF
  tiles = realm_map.tiles
  if start == goal:
    return (0, 0)
  if (start, goal) in realm_map.pathfinding_cache:
    return realm_map.pathfinding_cache[(start, goal)]
  initial_goal = goal
  pq = [(0, start)]

  backtrace = {}
  cost = {start: 0}

  closestPos = start
  closestHeuristic = l1(start, goal)
  closestCost = closestHeuristic

  while pq:
    # Use approximate solution if budget exhausted
    cutoff -= 1
    if cutoff <= 0:
      if goal not in backtrace:
        goal = closestPos
      break

    priority, cur = heapq.heappop(pq)

    if cur == goal:
      break

    for nxt in adjacentPos(cur):
      if not in_bounds(*nxt, tiles.shape):
        continue

      newCost = cost[cur] + 1
      if nxt not in cost or newCost < cost[nxt]:
        cost[nxt] = newCost
        heuristic = lInfty(goal, nxt)
        priority = newCost + heuristic

        # Compute approximate solution
        if heuristic < closestHeuristic or (
            heuristic == closestHeuristic and priority < closestCost):
          closestPos = nxt
          closestHeuristic = heuristic
          closestCost = priority

        heapq.heappush(pq, (priority, nxt))
        backtrace[nxt] = cur

  while goal in backtrace and backtrace[goal] != start:
    gr, gc = goal
    goal = backtrace[goal]
    sr, sc = goal
    realm_map.pathfinding_cache[(goal, initial_goal)] = (gr - sr, gc - sc)

  sr, sc = start
  gr, gc = goal
  realm_map.pathfinding_cache[(start, initial_goal)] = (gr - sr, gc - sc)
  return (gr - sr, gc - sc)
# End A*

# Adjacency functions
def adjacentDeltas():
  return [(-1, 0), (1, 0), (0, 1), (0, -1)]


def l1Deltas(s):
  rets = []
  for r in range(-s, s + 1):
    for c in range(-s, s + 1):
      rets.append((r, c))
  return rets


def posSum(pos1, pos2):
  return pos1[0] + pos2[0], pos1[1] + pos2[1]


def adjacentEmptyPos(env, pos):
  return [p for p in adjacentPos(pos)
          if in_bounds(*p, env.size)]


def adjacentTiles(env, pos):
  return [env.tiles[p] for p in adjacentPos(pos)
          if in_bounds(*p, env.size)]


def adjacentMats(tiles, pos):
  return [type(tiles[p].state) for p in adjacentPos(pos)
          if in_bounds(*p, tiles.shape)]


def adjacencyDelMatPairs(env, pos):
  return zip(adjacentDeltas(), adjacentMats(env.tiles, pos))
### End###



import os
import logging

import numpy as np
import vec_noise
from imageio.v2 import imread, imsave
from scipy import stats

from nmmo import material


def sharp(noise):
  '''Exponential noise sharpener for perlin ridges'''
  return 2 * (0.5 - abs(0.5 - noise))

class Save:
  '''Save utility for map files'''
  @staticmethod
  def render(mats, lookup, path):
    '''Render tiles to png'''
    images = [[lookup[e] for e in l] for l in mats]
    image = np.vstack([np.hstack(e) for e in images])
    imsave(path, image)

  @staticmethod
  def fractal(terrain, path):
    '''Render raw noise fractal to png'''
    frac = (256*terrain).astype(np.uint8)
    imsave(path, frac)

  @staticmethod
  def as_numpy(mats, path):
    '''Save map to .npy'''
    path = os.path.join(path, 'map.npy')
    np.save(path, mats.astype(int))

# pylint: disable=E1101:no-member
# Terrain uses setattr()
class Terrain:
  '''Terrain material class; populated at runtime'''
  @staticmethod
  def generate_terrain(config, map_id, interpolaters):
    center      = config.MAP_CENTER
    border      = config.MAP_BORDER
    size        = config.MAP_SIZE
    frequency   = config.TERRAIN_FREQUENCY
    offset      = config.TERRAIN_FREQUENCY_OFFSET
    octaves     = center // config.TERRAIN_TILES_PER_OCTAVE

    #Compute a unique seed based on map index
    #Flip seed used to ensure train/eval maps are different
    seed = map_id + 1
    if config.TERRAIN_FLIP_SEED:
      seed = -seed

    #Log interpolation factor
    if not interpolaters:
      interpolaters = np.logspace(config.TERRAIN_LOG_INTERPOLATE_MIN,
              config.TERRAIN_LOG_INTERPOLATE_MAX, config.MAP_N)

    interpolate = interpolaters[map_id]

    #Data buffers
    val   = np.zeros((size, size, octaves))
    scale = np.zeros((size, size, octaves))
    s     = np.arange(size)
    X, Y  = np.meshgrid(s, s)

    #Compute noise over logscaled octaves
    start = frequency
    end   = min(start, start - np.log2(center) + offset)
    for idx, freq in enumerate(np.logspace(start, end, octaves, base=2)):
      val[:, :, idx] = vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y)

    #Compute L1 distance
    x      = np.abs(np.arange(size) - size//2)
    X, Y   = np.meshgrid(x, x)
    data   = np.stack((X, Y), -1)
    l1     = np.max(abs(data), -1)

    #Interpolation Weights
    rrange = np.linspace(-1, 1, 2*octaves-1)
    pdf    = stats.norm.pdf(rrange, 0, interpolate)
    pdf    = pdf / max(pdf)
    high   = center / 2
    delta  = high / octaves

    #Compute perlin mask
    noise  = np.zeros((size, size))
    X, Y   = np.meshgrid(s, s)
    expand = int(np.log2(center)) - 2
    for idx, octave in enumerate(range(expand, 1, -1)):
      freq, mag = 1 / 2**octave, 1 / 2**idx
      noise    += mag * vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y)

    noise -= np.min(noise)
    noise = octaves * noise / np.max(noise) - 1e-12
    noise = noise.astype(int)

    #Compute L1 and Perlin scale factor
    for i in range(octaves):
      start             = octaves - i - 1
      scale[l1 <= high] = np.arange(start, start + octaves)
      high             -= delta

    start   = noise - 1
    l1_scale = np.clip(l1, 0, size//2 - border - 2)
    l1_scale = l1_scale / np.max(l1_scale)
    for i in range(octaves):
      idxs           = l1_scale*scale[:, :, i] + (1-l1_scale)*(start + i)
      scale[:, :, i] = pdf[idxs.astype(int)]

    #Blend octaves
    std = np.std(val)
    val = val / std
    val = scale * val
    val = np.sum(scale * val, -1)
    val = std * val / np.std(val)
    val = 0.5 + np.clip(val, -1, 1)/2

    #Threshold to materials
    matl = np.zeros((size, size), dtype=object)
    for y in range(size):
      for x in range(size):
        v = val[y, x]
        if v <= config.TERRAIN_WATER:
          mat = Terrain.WATER
        elif v <= config.TERRAIN_GRASS:
          mat = Terrain.GRASS
        elif v <= config.TERRAIN_FOILAGE:
          mat = Terrain.FOILAGE
        else:
          mat = Terrain.STONE
        matl[y, x] = mat

    # Void and grass border
    matl[l1 > size/2 - border]   = Terrain.VOID
    matl[l1 == size//2 - border] = Terrain.GRASS

    edge  = l1 == size//2 - border - 1
    stone = (matl == Terrain.STONE) | (matl == Terrain.WATER)
    matl[edge & stone] = Terrain.FOILAGE

    return val, matl, interpolaters

def place_fish(tiles, np_random):
  placed = False
  allow = {Terrain.GRASS}

  water_loc = np.where(tiles == Terrain.WATER)
  water_loc = list(zip(water_loc[0], water_loc[1]))
  np_random.shuffle(water_loc)

  for r, c in water_loc:
    if tiles[r-1, c] in allow or tiles[r+1, c] in allow or \
       tiles[r, c-1] in allow or tiles[r, c+1] in allow:
      tiles[r, c] = Terrain.FISH
      placed = True
      break

  if not placed:
    raise RuntimeError('Could not find the water tile to place fish.')

def uniform(config, tiles, mat, mmin, mmax, np_random):
  r = np_random.integers(mmin, mmax)
  c = np_random.integers(mmin, mmax)

  if tiles[r, c] not in {Terrain.GRASS}:
    uniform(config, tiles, mat, mmin, mmax, np_random)
  else:
    tiles[r, c] = mat

def cluster(config, tiles, mat, mmin, mmax, np_random):
  mmin = mmin + 1
  mmax = mmax - 1

  r = np_random.integers(mmin, mmax)
  c = np_random.integers(mmin, mmax)

  matls = {Terrain.GRASS}
  if tiles[r, c] not in matls:
    cluster(config, tiles, mat, mmin-1, mmax+1, np_random)
    return

  tiles[r, c] = mat
  if tiles[r-1, c] in matls:
    tiles[r-1, c] = mat
  if tiles[r+1, c] in matls:
    tiles[r+1, c] = mat
  if tiles[r, c-1] in matls:
    tiles[r, c-1] = mat
  if tiles[r, c+1] in matls:
    tiles[r, c+1] = mat

def spawn_profession_resources(config, tiles, np_random=None):
  if np_random is None:
    np_random = np.random

  mmin = config.MAP_BORDER + 1
  mmax = config.MAP_SIZE - config.MAP_BORDER - 1

  for _ in range(config.PROGRESSION_SPAWN_CLUSTERS):
    cluster(config, tiles, Terrain.ORE, mmin, mmax, np_random)
    cluster(config, tiles, Terrain.TREE, mmin, mmax, np_random)
    cluster(config, tiles, Terrain.CRYSTAL, mmin, mmax, np_random)

  for _ in range(config.PROGRESSION_SPAWN_UNIFORMS):
    uniform(config, tiles, Terrain.HERB, mmin, mmax, np_random)
    place_fish(tiles, np_random)

class MapGenerator:
  '''Procedural map generation'''
  def __init__(self, config):
    self.config = config
    self.load_textures()
    self.interpolaters = None

  def load_textures(self):
    '''Called during setup; loads and resizes tile pngs'''
    lookup = {}
    path   = self.config.PATH_TILE
    scale  = self.config.MAP_PREVIEW_DOWNSCALE
    for mat in material.All:
      key = mat.tex
      tex = imread(path.format(key))
      lookup[mat.index] = tex[:, :, :3][::scale, ::scale]
      setattr(Terrain, key.upper(), mat.index)
    self.textures = lookup

  def generate_all_maps(self, np_random=None):
    '''Generates NMAPS maps according to generate_map

    Provides additional utilities for saving to .npy and rendering png previews'''

    config = self.config

    #Only generate if maps are not cached
    path_maps = os.path.join(config.PATH_CWD, config.PATH_MAPS)
    os.makedirs(path_maps, exist_ok=True)

    existing_maps = set(map_dir + '/map.npy' for map_dir in os.listdir(path_maps))
    if not config.MAP_FORCE_GENERATION and existing_maps:
      required_maps = {
        f'map{idx}/map.npy' for idx in range(1, config.MAP_N+1)
      }
      missing = required_maps - existing_maps
      if not missing:
        return

    if __debug__:
      logging.info('Generating %s maps', str(config.MAP_N))

    for idx in range(config.MAP_N):
      path = path_maps + '/map' + str(idx+1)
      os.makedirs(path, exist_ok=True)

      terrain, tiles = self.generate_map(idx, np_random)

      #Save/render
      Save.as_numpy(tiles, path)
      if config.MAP_GENERATE_PREVIEWS:
        b = config.MAP_BORDER
        tiles = [e[b:-b+1] for e in tiles][b:-b+1]
        Save.fractal(terrain, path+'/fractal.png')
        Save.render(tiles, self.textures, path+'/map.png')

  def generate_map(self, idx, np_random=None):
    '''Generate a single map

    The default method is a relatively complex multiscale perlin noise method.
    This is not just standard multioctave noise -- we are seeding multioctave noise
    itself with perlin noise to create localized deviations in scale, plus additional
    biasing to decrease terrain frequency towards the center of the map

    We found that this creates more visually interesting terrain and more deviation in
    required planning horizon across different parts of the map. This is by no means a
    gold-standard: you are free to override this method and create customized terrain
    generation more suitable for your application. Simply pass MAP_GENERATOR=YourMapGenClass
    as a config argument.'''
    config = self.config
    if config.TERRAIN_SYSTEM_ENABLED:
      if not hasattr(self, 'interpolaters'):
        self.interpolaters = None
      terrain, tiles, _ = Terrain.generate_terrain(config, idx, self.interpolaters)
    else:
      size    = config.MAP_SIZE
      terrain = np.zeros((size, size))
      tiles   = np.zeros((size, size), dtype=object)

      for r in range(size):
        for c in range(size):
          linf = max(abs(r - size//2), abs(c - size//2))
          if linf <= size//2 - config.MAP_BORDER:
            tiles[r, c] = Terrain.GRASS
          else:
            tiles[r, c] = Terrain.VOID

    if config.PROFESSION_SYSTEM_ENABLED:
      spawn_profession_resources(config, tiles, np_random)

    return terrain, tiles



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict

import nmmo
from nmmo.core.log_helper import LogHelper
from nmmo.core.map import Map
from nmmo.core.tile import TileState
from nmmo.core.action import Action, Buy
from nmmo.entity.entity import EntityState
from nmmo.entity.entity_manager import NPCManager, PlayerManager
from nmmo.datastore.numpy_datastore import NumpyDatastore
from nmmo.systems.exchange import Exchange
from nmmo.systems.item import Item, ItemState
from nmmo.lib.event_log import EventLogger, EventState
from nmmo.render.replay_helper import ReplayHelper

def prioritized(entities: Dict, merged: Dict):
  """Sort actions into merged according to priority"""
  for idx, actions in entities.items():
    for atn, args in actions.items():
      merged[atn.priority].append((idx, (atn, args.values())))
  return merged


class Realm:
  """Top-level world object"""

  def __init__(self, config, np_random):
    self.config = config
    self._np_random = np_random # rng
    assert isinstance(
        config, nmmo.config.Config
    ), f"Config {config} is not a config instance (did you pass the class?)"

    Action.hook(config)

    # Generate maps if they do not exist
    # NOTE: Map generation interferes with determinism.
    #   To ensure determinism, provide seed to env.reset()
    config.MAP_GENERATOR(config).generate_all_maps(self._np_random)

    self.datastore = NumpyDatastore()
    for s in [TileState, EntityState, ItemState, EventState]:
      self.datastore.register_object_type(s._name, s.State.num_attributes)

    self.tick = None # to use as a "reset" checker
    self.exchange = None

    # Load the world file
    self.map = Map(config, self, self._np_random)

    self.log_helper = LogHelper.create(self)
    self.event_log = EventLogger(self)

    # Entity handlers
    self.players = PlayerManager(self, self._np_random)
    self.npcs = NPCManager(self, self._np_random)

    # Global item registry
    self.items = {}

    # Replay helper
    self._replay_helper = None

    # Initialize actions
    nmmo.Action.init(config)

  def reset(self, np_random, map_id: int = None):
    """Reset the environment and load the specified map

    Args:
        idx: Map index to load
    """
    self._np_random = np_random
    self.log_helper.reset()
    self.event_log.reset()

    map_id = map_id or self._np_random.integers(self.config.MAP_N) + 1
    self.map.reset(map_id, self._np_random)
    self.tick = 0

    # EntityState and ItemState tables must be empty after players/npcs.reset()
    self.players.reset(self._np_random)
    self.npcs.reset(self._np_random)
    assert EntityState.State.table(self.datastore).is_empty(), \
        "EntityState table is not empty"
    # TODO: fix the item leak, then uncomment the below -- print out the table?
    # assert ItemState.State.table(self.datastore).is_empty(), \
    #     "ItemState table is not empty"

    # DataStore id allocator must be reset to be deterministic
    EntityState.State.table(self.datastore).reset()
    ItemState.State.table(self.datastore).reset()

    self.players.spawn()
    self.npcs.spawn()

    # Global item exchange
    self.exchange = Exchange(self)

    # Global item registry
    Item.INSTANCE_ID = 0
    self.items = {}

    if self._replay_helper is not None:
      self._replay_helper.reset()

  def packet(self):
    """Client packet"""
    return {
      "environment": self.map.repr,
      "border": self.config.MAP_BORDER,
      "size": self.config.MAP_SIZE,
      "resource": self.map.packet,
      "player": self.players.packet,
      "npc": self.npcs.packet,
      "market": self.exchange.packet,
  }

  @property
  def num_players(self):
    """Number of player agents"""
    return len(self.players.entities)

  def entity(self, ent_id):
    e = self.entity_or_none(ent_id)
    assert e is not None, f"Entity {ent_id} does not exist"
    return e

  def entity_or_none(self, ent_id):
    if ent_id is None:
      return None

    """Get entity by ID"""
    if ent_id < 0:
      return self.npcs.get(ent_id)

    return self.players.get(ent_id)

  def step(self, actions):
    """Run game logic for one tick

    Args:
        actions: Dict of agent actions

    Returns:
        dead: List of dead agents
    """
    # Prioritize actions
    npc_actions = self.npcs.actions(self)
    merged = defaultdict(list)
    prioritized(actions, merged)
    prioritized(npc_actions, merged)

    # Update entities and perform actions
    self.players.update(actions)
    self.npcs.update(npc_actions)

    # Execute actions -- CHECK ME the below priority
    #  - 10: Use - equip ammo, restore HP, etc.
    #  - 20: Buy - exchange while sellers, items, buyers are all intact
    #  - 30: Give, GiveGold - transfer while both are alive and at the same tile
    #  - 40: Destroy - use with SELL/GIVE, if not gone, destroy and recover space
    #  - 50: Attack
    #  - 60: Move
    #  - 70: Sell - to guarantee the listed items are available to buy
    #  - 99: Comm

    for priority in sorted(merged):
      # TODO: we should be randomizing these, otherwise the lower ID agents
      # will always go first. --> ONLY SHUFFLE BUY
      if priority == Buy.priority:
        self._np_random.shuffle(merged[priority])

      # CHECK ME: do we need this line?
      # ent_id, (atn, args) = merged[priority][0]
      for ent_id, (atn, args) in merged[priority]:
        ent = self.entity(ent_id)
        if ent.alive:
          atn.call(self, ent, *args)
    dead = self.players.cull()
    self.npcs.cull()

    # Update map
    self.map.step()
    self.exchange.step(self.tick)
    self.log_helper.update(dead)
    self.event_log.update()
    if self._replay_helper is not None:
      self._replay_helper.update()

    self.tick += 1

    return dead

  def log_milestone(self, category: str, value: float, message: str = None, tags: Dict = None):
    self.log_helper.log_milestone(category, value)
    self.log_helper.log_event(category, value)

    if self.config.LOG_VERBOSE:
      # TODO: more general handling of tags, if necessary
      if tags and 'player_id' in tags:
        logging.info("Milestone (Player %d): %s %s %s", tags['player_id'], category, value, message)
      else:
        logging.info("Milestone: %s %s %s", category, value, message)

  def record_replay(self, replay_helper: ReplayHelper) -> ReplayHelper:
    self._replay_helper = replay_helper
    self._replay_helper.set_realm(self)

    return replay_helper



from __future__ import annotations
from typing import Dict, List
from nmmo.datastore.id_allocator import IdAllocator

"""
This code defines a data storage system that allows for the
creation, manipulation, and querying of records.

The DataTable class serves as the foundation for the data
storage, providing methods for updating and retrieving data,
as well as filtering and querying records.

The DatastoreRecord class represents a single record within
a table and provides a simple interface for interacting with
the data. The Datastore class serves as the main entry point
for the data storage system, allowing for the creation and
management of tables and records.

The implementation of the DataTable class is left to the
developer, but the DatastoreRecord and Datastore classes
should be sufficient for most use cases.

See numpy_datastore.py for an implementation.
"""
class DataTable:
  def __init__(self, num_columns: int):
    self._num_columns = num_columns
    self._id_allocator = IdAllocator(100)

  def reset(self):
    self._id_allocator = IdAllocator(100)

  def update(self, row_id: int, col: int, value):
    raise NotImplementedError

  def get(self, ids: List[id]):
    raise NotImplementedError

  def where_in(self, col: int, values: List):
    raise NotImplementedError

  def where_eq(self, col: str, value):
    raise NotImplementedError

  def where_neq(self, col: str, value):
    raise NotImplementedError

  def window(self, row_idx: int, col_idx: int, row: int, col: int, radius: int):
    raise NotImplementedError

  def remove_row(self, row_id: int):
    raise NotImplementedError

  def add_row(self) -> int:
    raise NotImplementedError

  def is_empty(self) -> bool:
    raise NotImplementedError

class DatastoreRecord:
  def __init__(self, datastore, table: DataTable, row_id: int) -> None:
    self.datastore = datastore
    self.table = table
    self.id = row_id

  def update(self, col: int, value):
    self.table.update(self.id, col, value)

  def get(self, col: int):
    return self.table.get(self.id)[col]

  def delete(self):
    self.table.remove_row(self.id)

class Datastore:
  def __init__(self) -> None:
    self._tables: Dict[str, DataTable] = {}

  def register_object_type(self, object_type: str, num_colums: int):
    if object_type not in self._tables:
      self._tables[object_type] = self._create_table(num_colums)

  def create_record(self, object_type: str) -> DatastoreRecord:
    table = self._tables[object_type]
    row_id = table.add_row()
    return DatastoreRecord(self, table, row_id)

  def table(self, object_type: str) -> DataTable:
    return self._tables[object_type]

  def _create_table(self, num_columns: int) -> DataTable:
    raise NotImplementedError



# pylint: disable=bad-builtin, unused-variable
import psutil

import nmmo

def test_memory_usage():
  env = nmmo.Env()
  process = psutil.Process()
  print("memory", process.memory_info().rss)

if __name__ == '__main__':
  test_memory_usage()



import os
import logging
import numpy as np
from ordered_set import OrderedSet

from nmmo.core.tile import Tile
from nmmo.lib import material


class Map:
  '''Map object representing a list of tiles

  Also tracks a sparse list of tile updates
  '''
  def __init__(self, config, realm, np_random):
    self.config = config
    self._repr  = None
    self.realm  = realm
    self.update_list = None
    self.pathfinding_cache = {} # Avoid recalculating A*, paths don't move

    sz          = config.MAP_SIZE
    self.tiles  = np.zeros((sz, sz), dtype=object)
    self.habitable_tiles = np.zeros((sz,sz))

    for r in range(sz):
      for c in range(sz):
        self.tiles[r, c] = Tile(realm, r, c, np_random)

    self.dist_border_center = config.MAP_CENTER // 2
    self.center_coord = (config.MAP_BORDER + self.dist_border_center,
                         config.MAP_BORDER + self.dist_border_center)

  @property
  def packet(self):
    '''Packet of degenerate resource states'''
    missing_resources = []
    for e in self.update_list:
      missing_resources.append(e.pos)
    return missing_resources

  @property
  def repr(self):
    '''Flat matrix of tile material indices'''
    if not self._repr:
      self._repr = [[t.material.index for t in row] for row in self.tiles]

    return self._repr

  def reset(self, map_id, np_random):
    '''Reuse the current tile objects to load a new map'''
    config = self.config
    self.update_list = OrderedSet() # critical for determinism

    path_map_suffix = config.PATH_MAP_SUFFIX.format(map_id)
    f_path = os.path.join(config.PATH_CWD, config.PATH_MAPS, path_map_suffix)

    try:
      map_file = np.load(f_path)
    except FileNotFoundError:
      logging.error('Maps not found')
      raise

    materials = {mat.index: mat for mat in material.All}
    r, c = 0, 0
    for r, row in enumerate(map_file):
      for c, idx in enumerate(row):
        mat  = materials[idx]
        tile = self.tiles[r, c]
        tile.reset(mat, config, np_random)
        self.habitable_tiles[r, c] = tile.habitable

    assert c == config.MAP_SIZE - 1
    assert r == config.MAP_SIZE - 1

    self._repr = None

  def step(self):
    '''Evaluate updatable tiles'''
    self.realm.log_milestone('Resource_Depleted', len(self.update_list),
        f'RESOURCE: Depleted {len(self.update_list)} resource tiles')

    for e in self.update_list.copy():
      if not e.depleted:
        self.update_list.remove(e)
      e.step()

  def harvest(self, r, c, deplete=True):
    '''Called by actions that harvest a resource tile'''

    if deplete:
      self.update_list.add(self.tiles[r, c])

    return self.tiles[r, c].harvest(deplete)

  def is_valid_pos(self, row, col):
    '''Check if a position is valid'''
    return 0 <= row < self.config.MAP_SIZE and 0 <= col < self.config.MAP_SIZE



import unittest
import numpy as np

import nmmo
from nmmo.entity.entity import Entity, EntityState
from nmmo.datastore.numpy_datastore import NumpyDatastore

class MockRealm:
  def __init__(self):
    self.config = nmmo.config.Default()
    self.config.PLAYERS = range(100)
    self.datastore = NumpyDatastore()
    self.datastore.register_object_type("Entity", EntityState.State.num_attributes)
    self._np_random = np.random

# pylint: disable=no-member
class TestEntity(unittest.TestCase):
  def test_entity(self):
    realm = MockRealm()
    entity_id = 123
    entity = Entity(realm, (10,20), entity_id, "name")

    self.assertEqual(entity.id.val, entity_id)
    self.assertEqual(entity.row.val, 10)
    self.assertEqual(entity.col.val, 20)
    self.assertEqual(entity.damage.val, 0)
    self.assertEqual(entity.time_alive.val, 0)
    self.assertEqual(entity.freeze.val, 0)
    self.assertEqual(entity.item_level.val, 0)
    self.assertEqual(entity.attacker_id.val, 0)
    self.assertEqual(entity.message.val, 0)
    self.assertEqual(entity.gold.val, 0)
    self.assertEqual(entity.health.val, realm.config.PLAYER_BASE_HEALTH)
    self.assertEqual(entity.food.val, realm.config.RESOURCE_BASE)
    self.assertEqual(entity.water.val, realm.config.RESOURCE_BASE)
    self.assertEqual(entity.melee_level.val, 0)
    self.assertEqual(entity.range_level.val, 0)
    self.assertEqual(entity.mage_level.val, 0)
    self.assertEqual(entity.fishing_level.val, 0)
    self.assertEqual(entity.herbalism_level.val, 0)
    self.assertEqual(entity.prospecting_level.val, 0)
    self.assertEqual(entity.carving_level.val, 0)
    self.assertEqual(entity.alchemy_level.val, 0)

  def test_query_by_ids(self):
    realm = MockRealm()
    entity_id = 123
    entity = Entity(realm, (10,20), entity_id, "name")

    entities = EntityState.Query.by_ids(realm.datastore, [entity_id])
    self.assertEqual(len(entities), 1)
    self.assertEqual(entities[0][Entity.State.attr_name_to_col["id"]], entity_id)
    self.assertEqual(entities[0][Entity.State.attr_name_to_col["row"]], 10)
    self.assertEqual(entities[0][Entity.State.attr_name_to_col["col"]], 20)

    entity.food.update(11)
    e_row = EntityState.Query.by_id(realm.datastore, entity_id)
    self.assertEqual(e_row[Entity.State.attr_name_to_col["food"]], 11)


if __name__ == '__main__':
  unittest.main()
